//#define _ANSI_C_SOURCE

#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include "statistic.hh"

#include <iostream.h>
#include <fstream.h>

#include "sys_time.hh"
 
/***********************************************************************
**                       Table                                        **
***********************************************************************/

Table::Table(unsigned int dim)
{
  if(dim == 0)
    dim = 1;

  int n =(int) ceil(log((double)dim)/log(2.0));
  if(n > 20)  // max 1M 
    n = 20; 
  if(n < 14)  // min 16K 
    n = 14;

  tables_pow2dim = n - 5;
  tables_dim =(int) 0x01 << tables_pow2dim;
  cout << "TABLE: sub tables dim = " << tables_dim << '\n';
  
  for(int i = 0;i<32;i++)
    tables_array[i] = NULL;  
}

Table::Table(Table& T)
{
  tables_pow2dim = T.tables_pow2dim;
  tables_dim     = T.tables_dim;
  
  for(int i = 0;i<32;i++)
    if(T.tables_array[i] != NULL)
      {
	tables_array[i] = new pair[tables_dim];
	bcopy((char*)T.tables_array[i],(char*)tables_array[i],sizeof(pair)*tables_dim);
      }
    else
      tables_array[i] = NULL;
}
  
Table::~Table()
{
  for(int i = 0;i<32;i++)
    if(tables_array[i])
      delete tables_array[i];
}

pair* Table::operator[](unsigned int index)
{
  if(index >= (32*tables_dim))
    {
      cerr << "TABLE: requested elements with out of range index (" << index << 
	"). Exit immediately\n";
      exit(-1);
    }

  //unsigned int hi_offset  = index & ~((0x01 << tables_pow2dim)-1);
  unsigned int hi_offset  = (index & ~((0x01 << tables_pow2dim)-1)) >> tables_pow2dim;
  unsigned int low_offset = index & ((0x01 << tables_pow2dim)-1);

  if(tables_array[hi_offset] == NULL)
    {
      //cout << "TABLE: new subtable\n";
      tables_array[hi_offset] = new pair[tables_dim];
      bzero((char*)tables_array[hi_offset],sizeof(pair)*tables_dim);
    }
  
  //cout << "index = " << index << '\n';
  //cout << "hi = " << hi_offset << ", low = " << low_offset << '\n';
  return (tables_array[hi_offset] + low_offset);
}


/***********************************************************************
**                       Statistic                                    **
***********************************************************************/

sys_time Statistic::start_time(GetActualSystemTime());

Statistic::Statistic(char* _name)
: data(131072)
{
  elements = 0;
  if(_name != NULL)    
    strcpy(name,_name);
  else
    strcpy(name,"unnamed");
}

Statistic::Statistic(Statistic& S)
: data(S.data)
{
  elements = S.elements;
  strcpy(name,S.name);
}

Statistic::~Statistic()
{
  ;
}

void Statistic::WriteToFile(char* filename)
{
  if(filename == NULL)
    filename = name;

  cout << "STAT: Write data for " << filename << '\n';

  pair* p;
  //double x,y;
  ofstream  file(filename);
  for(int i=0;i<elements;i++)
    {
      //cout << "STAT: write data\n";
      p = data[i];
      //cout << p->X << "\t\t" << p->Y << '\n';
      file << p->X << "\t\t" << p->Y << '\n';
    }

  char  filename_op[256];
  strcpy(filename_op,filename);
  strcat(filename_op,".opnet");
  ofstream  file_op(filename_op);
  file_op << "trace_count = 1\n";
  file_op << "ordinate: time\n"; 
  file_op << "abscissa: " << filename << '\n'; 
  file_op << "length = " << elements << '\n';
  file_op << "number of values = " << elements << '\n';
  for(i=0;i<elements;i++)
    {
      p = data[i];
      file_op << p->X << " -> " << p->Y << '\n';
    }
}


void Statistic::ReadFromFile(char* filename)
{
  ;
}

const char* Statistic::Name()
{
  return (const char*) name;
}

unsigned int Statistic::Elements()
{
  return elements;
}

void Statistic::Add()
{
  //pair   p;
  sys_time  t = GetActualSystemTime() - start_time;
  //cout << t << '\n';
  double db = (double) t * 1e-9;
  data[elements]->X = db;
  data[elements]->Y = db;
  //data[elements] = p;
  elements++;
}

void Statistic::Add(double db)
{
  //pair p;
  sys_time  t = GetActualSystemTime() - start_time;
  data[elements]->X = (double) t * 1e-9;
  data[elements]->Y = db;
  //data[elements] = p;
  elements++;
}

void Statistic::Read(unsigned int index,double* pvx,double* pvy)
{
  if(index > elements)
    {
      *pvx = 0.0;
      *pvy = 0.0;
    }
  else
    {
      *pvx = data[index]->X;
      *pvy = data[index]->Y;
    }
}

double Statistic::Max()
{
  double max = - DBL_MAX;

  for(int i=0;i<elements;i++)
    if(data[i]->Y > max)
      max = data[i]->Y;
  
  return max;
}

double Statistic::Min()
{
  double min = DBL_MAX;

  for(int i=0;i<elements;i++)
    if(data[i]->Y < min)
      min = data[i]->Y;
  
  return min;
}


double Statistic::Sample_mean()
{
  double mean = 0.0;
  
  for(int i=0;i<elements;i++)
  {
    mean += data[i]->Y;
    //cout << mean << '\t' << data[i].Y << '\n';
  }
  
  return mean/elements;
}

/*
void Statistic::Delta()
{
  //cout << "Delta:\n";

  pair*   p1;
  pair*   p2;

  if(elements >= 2)
    {
      elements--;
      for(int i=0;i<elements;i++)
	{
	  //cout << "Delta: " << i << '\n';
	  p1 = data[i];
	  p2 = data[i+1];

	  p1->X = p2->X;
	  //data[i].Y = data[i+1].Y - data[i-1].Y;
	  p1->Y = p2->Y - p1->Y;
	}
    }
  //cout << "Delta: end\n";
}
*/

/*
Statistic* Delta(Statistic& S1)
{
  Statistic* pS2 = new Statistic("tmp");
  
  if(S1.elements < 2)
    return pS2;

  pair*   p11;
  pair*   p12;
  pair*   p2;

  pS2->elements = S1.elements - 1;
  for(int i=0;i<pS2->elements;i++)
    {
      //cout << "Delta: " << i << '\n';

      p11 = (S1.data)[i];
      p12 = (S1.data)[i+1];
      p2  = (pS2->data)[i];
      
      p2->X = p12->X;
      //cout << p2->X << '\n';
      p2->Y = p12->Y - p11->Y;
      //cout << p12->Y << '\t' << p11->Y << '\n';
    }

  pair* p;
  for(i=0;i<pS2->elements;i++)
    {
      p = (pS2->data)[i];
      cout << p->X << "\t\t" << p->Y << '\n';
    }

  return pS2;
}
*/

Statistic Delta(Statistic& S1)
{
  Statistic S2("tmp");
  
  if(S1.elements < 2)
    return S2;

  pair*   p11;
  pair*   p12;
  pair*   p2;

  S2.elements = S1.elements - 1;
  for(int i=0;i<S2.elements;i++)
    {
      p11 = (S1.data)[i];
      p12 = (S1.data)[i+1];
      p2  = (S2.data)[i];
      
      p2->X = p12->X;
      //cout << p2->X << '\n';
      p2->Y = p12->Y - p11->Y;
      //cout << p12->Y << '\t' << p11->Y << '\n';
    }

  return S2;
}

Statistic operator-(Statistic& S1,Statistic& S2)
{
  Statistic S3("tmp");
  
  if(S1.elements != S2.elements)
    return S3;

  pair* p1;
  pair* p2;
  pair* p3;
  S3.elements = S1.elements;
  for(int i=0;i<S3.elements;i++)
    {
      p3 = S3.data[i];
      p1 = S1.data[i];
      p2 = S2.data[i];
      p3->X = p1->X;
      p3->Y = p1->Y - p2->Y;
    }
  return S3;
}

Statistic Distribution(Statistic& S1)
{
  Statistic S2("tmp");

  if(S1.elements == 0)
    return S2;

  pair* p1;
  pair* p2;
  S2.elements            = 2 * S1.elements;
  double        interval = S1.Max() - S1.Min();
  double        delta    = interval / S2.elements;
  double        min      = S1.Min(); 
  unsigned int  frequency_table[S2.elements];
  unsigned int  index    = 0;

  for(int i=0;i<S2.elements;i++)
    frequency_table[i] = 0;

  for(i=0;i<S1.elements;i++)
    {
      p1 = S1.data[i];
      index =(unsigned int) floor((p1->Y - min) / delta);
      frequency_table[index]++;
    }
  
  for(i=0;i<S2.elements;i++)
    {
      p2 = (S2.data)[i];
      p2->X = min + delta * (double) i;
      p2->Y = ((double) frequency_table[i]) / S1.elements;
    }

  return S2;
}

Statistic Probability(Statistic& S1)
{
  Statistic S2(Distribution(S1));

  if(S1.elements == 0)
    return S2;
  
  double s = 0.0;
  pair*  p2;
  for(int i=0;i<S2.elements;i++)
    {
      p2 = (S2.data)[i];
      s += p2->Y;
      p2->Y = s;
    }
  return S2;
}
  
/***********************************************************************
**                           Statistic_Manager                        **
***********************************************************************/
						 
Statistic_Manager::Statistic_Manager()
: elements(0)
{
  int i = 0;

  for(;i<max_stat_number;i++)
    {
      pstat_array[i] = NULL;
      strcpy(stat_name_array[i],"");
    }
}

Statistic_Manager::~Statistic_Manager()
{
  ;
}


Statistic* Statistic_Manager::QueryNewStat(char* name)
{
  if(elements > max_stat_number)
    return NULL;

  Statistic* pstat = new Statistic(name);
  if(pstat == NULL)
    return NULL;

  for(int i = 0;i<max_stat_number;i++)
    if(pstat_array[i] == NULL)
      {
	pstat_array[i] = pstat;
	strcpy(stat_name_array[i],name);
	return pstat;
      }

  delete pstat;
  return NULL;
}

Statistic* Statistic_Manager::QueryByName(char* name)
{
  for(int i = 0;i<max_stat_number;i++)
    if(strcmp(stat_name_array[i],name) == 0)
      return pstat_array[i];

  return NULL;
}


Statistic_Manager  The_Statistic_Manager;










